@@ -6,11 +6,6 @@ use POSIX qw(:sys_wait_h);
 use Fcntl qw(:seek :flock O_RDONLY O_RDWR O_TRUNC O_CREAT);
 use Digest::MD5;
 
-BEGIN {
-  eval { require File::Sync; };
-  $@ and *File::Sync::fsync = sub { 1 };
-}
-
 # usage
 #
 # use ddb;
@@ -24,14 +19,13 @@ BEGIN {
 # untie %db;
 
 # globals
-$VERSION	= '1.3.1';
+$version	= 1.1;
 $hash_size	= 16381; # default, or pass to tie after filename
 $sentinel	= 1;
 $empty_buf_size	= 256;
 $magic		= 0xDDB10000;
 $debug		= 0;
 $max_procs	= 10; # for test
-$show_step	= 100;
 $ptr_pos	= undef;
 
 # file format
@@ -81,12 +75,6 @@ DONE:
 sub STORE {
   my ($db, $key, $val) = @_;
 
-  unless (defined $val) {
-    # how else to make it undef?
-    $db->DELETE($key);
-    return undef;
-  }
-
   $db->lock_ex;
   my ($pos, $next_pos) = $db->find($key);
 
@@ -99,14 +87,14 @@ sub STORE {
 
     if ($old_val_len < $val_len) {
       my $rec = $db->pack_rec($key, $val, $next_pos);
-      $db->append_rec($rec);
+      my $new_pos = $db->append_rec($rec);
       my $old_rec_len = $db->rec_len($key_len, $old_val_len);
       $db->erase($pos, $old_rec_len);
     } else {
       $db->replace_val($key, $val, $pos, $next_pos, $old_val_len);
     }
   } else {
-    my $rec = $db->pack_rec($key, $val, 0);
+    my $rec = $db->pack_rec($key, $val, $next_pos);
     $db->append_rec($rec);
   }
 
@@ -128,11 +116,9 @@ sub DELETE {
 
   $db->seek($ptr_pos, SEEK_SET);
   $db->write_int($next_pos);
-  $db->sync;
 
   my $rec_len = $db->rec_len($key_len, $val_len);
   $db->erase($pos, $rec_len);
-  $db->sync;
 
 DONE:
   $db->lock_un;
@@ -149,9 +135,9 @@ sub CLEAR {
   $db->write_zero(4 * $db->{hash_size});
 
   my $pos = $db->tell;
+  $pos == $db->data_section or die;
   $db->truncate($pos);
 
-  $db->sync;
   $db->lock_un;
   ( )
 }
@@ -170,7 +156,6 @@ sub FIRSTKEY {
   my $db = shift;
   undef $db->{cur_hash};
   @{$db->{cur_keys}} = ( );
-  $$db{rec_count} = 0;
   $db->NEXTKEY
 }
 
@@ -183,7 +168,6 @@ sub TIEHASH {
     hash_size	=> $hash_size,
     cur_hash	=> undef,
     cur_keys	=> [ ],
-    rec_count	=> 0,
     lock_count	=> 0,
     lock_type	=> undef,
   }, $p;
@@ -214,7 +198,6 @@ sub TIEHASH {
   my $min_size = $db->data_section;
   $end_pos < $min_size and
     $db->die("file truncated, $end_pos / $min_size expected bytes");
-  $db->sync;
   $db->lock_un;
 
   $db
@@ -226,7 +209,6 @@ sub UNTIE {
   $db->{lock_count} = 0;
   $db->{lock_type} = undef;
   @{$db->{cur_keys}} = ( );
-  $db->{rec_count} = 0;
   $db->{cur_hash} = undef;
 
   close $db->{fh};
@@ -297,25 +279,6 @@ sub warn {
   warn "$0: $$db{filename}: $msg";
 }
 
-sub show_status {
-  my $db = shift;
-
-  defined($$db{cur_hash}) or return;
-  my $last_complete = int(100 * ($$db{cur_hash} - 1) / $$db{hash_size});
-  my $complete = int(100 * $$db{cur_hash} / $$db{hash_size});
-  $last_complete == $complete && $$db{rec_count} % $show_step and return;
-  my $nl = ($complete == 100) ? "  \r\n" : "  \r";
-  print STDERR "$0: $$db{rec_count} records, $complete% complete  $nl";
-}
-
-
-# file operations
-
-sub sync {
-  my $db = shift;
-  File::Sync::fsync($db->{fh}) or $db->warn('fsync failed');
-}
-
 sub tell {
   my $db = shift;
   sysseek $db->{fh}, 0, SEEK_CUR
@@ -396,7 +359,7 @@ sub read_key {
   my $key_len = $db->read_int;
 
   if (@_ > 1) {
-    $key_len < 0 || $pos + 9 + $key_len > $end_pos and
+    $key_len < 0 || $pos + $key_len > $end_pos and
       $db->die("key_len $key_len out of bounds");
   }
 
@@ -454,7 +417,7 @@ sub write_byte {
 
 sub write_sentinel {
   my $db = shift;
-  $db->write_byte($sentinel)
+  $db->write_byte(1)
 }
 
 sub write_int {
@@ -522,11 +485,8 @@ sub write_rec {
   }
 
   $db->write($rec);
-  $db->sync;
-
   $db->seek($ptr_pos, SEEK_SET);
   $db->write_int($pos);
-  $db->sync;
 }
 
 sub append_rec {
@@ -582,7 +542,6 @@ sub replace_val {
 
   $db->seek($ptr_pos, SEEK_SET);
   $db->write_int($pos);
-  $db->sync;
 
   $db->truncate($new_pos);
 
@@ -719,7 +678,6 @@ sub erase_panic {
 
   local $db->{cur_keys} = [ ];
   local $db->{cur_hash} = undef;
-  local $db->{rec_count} = 0;
   my $count = 0;
   
   while (1) {
@@ -743,8 +701,7 @@ sub erase_panic {
 # check each key right before returning it.
 
 sub next_pos {
-  my ($db, $status_cb) = @_;
-  $status_cb ||= sub { };
+  my $db = shift;
 
   $db->{cur_keys} ||= [ ];
   my $cur_keys = $db->{cur_keys};
@@ -753,17 +710,13 @@ sub next_pos {
   while (1) {
     while (defined(my $key = shift @$cur_keys)) {
       my ($pos, $next_pos) = $db->find($key);
-      if (defined($pos)) {
-        ++$db->{rec_count};
-        return ($pos, $key);
-      }
+      defined($pos) and return ($pos, $key);
       $debug and $db->warn("skipping unlinked cached record");
     }
 
     $db->{cur_hash} =
       defined($db->{cur_hash}) ?
       $db->{cur_hash} + 1 : 0;
-    $db->$status_cb;
     unless ($db->{cur_hash} < $db->{hash_size}) {
       undef $db->{cur_hash};
       return ( );
@@ -849,28 +802,13 @@ sub defrag {
     my $check_pos = $db->find($key);
     unless ($check_pos == $pos) {
       if ($check_val_hash == $val_hash) {
-        if (defined($check_pos)) {
-          # this can delete indexed data in a pathological case
-          # (a corrupted record with valid hash that overlaps indexed
-          # records, very unlikely by accident).  but it's doesn't
-          # have to scan the entire database like erase_panic.
-        
-          $db->warn("erasing unlinked record at $pos+$rec_len");
-          $empty_len += $db->erase($pos, $rec_len);
-        } else {
-          # this record is left over from an aborted delete or
-          # part of a chain after an erased corrupted record,
-          # so we relink it.
+        # this can delete indexed data in a pathological case
+        # (a corrupted record with valid hash that overlaps indexed
+        # records, very unlikely by accident).  but it's doesn't
+        # have to scan the entire database like erase_panic.
 
-          $db->warn("relinking unlinked record at $pos+$rec_len");
-
-          $db->seek($pos + 1, SEEK_SET);
-          $db->write_int(0);
-
-          $db->seek($ptr_pos, SEEK_SET);
-          $db->write_int($pos);
-          $db->sync;
-        }
+        $db->warn("erasing unlinked record at $pos+$rec_len");
+        $empty_len += $db->erase($pos, $rec_len);
       } else {
         $db->warn("val_hash mismatch at $pos+$rec_len");
         $empty_len += $db->erase_panic($pos, \&ep_status_cb);
@@ -881,22 +819,21 @@ sub defrag {
     $check_val_hash == $val_hash or
       $db->die("val_hash mismatch");
 
-    my $align = 3 - $empty_pos % 4;
-    $empty_pos += $align;
-    $empty_len -= $align;
-
-    if ($empty_len > 0) {
-      my $rec = $db->pack_rec($key, $val, $next_pos, $val_hash);
-      $db->move_rec($rec, $pos, $empty_pos);
-    } else {
-      # should never happen
-      $empty_len = 0;
+    if (my $align = 3 - $empty_pos % 4) {
+      $empty_pos += $align;
+      $empty_len -= $align;
     }
 
+    unless ($empty_len > 0) {
+      $empty_pos += $rec_len;
+      next;
+    }
+
+    my $rec = $db->pack_rec($key, $val, $next_pos, $val_hash);
+    $db->move_rec($rec, $pos, $empty_pos);
     $empty_pos += $rec_len;
   }
   
-  $db->sync;
   $db->$status_cb($end_pos, $end_pos);
   $db->lock_un;
 }
@@ -905,47 +842,36 @@ sub defrag {
 sub repair {
   my ($db, $status_cb) = @_;
   $status_cb ||= sub { };
+  $db->$status_cb(my $count = 0);
 
   local $debug = 1;
   local $db->{cur_keys} = [ ];
   local $db->{cur_hash} = undef;
-  local $db->{rec_count} = 0;
 
   $db->lock_ex;
 
   while (1) {
     $db->lock_sh;
-    my $pos = eval { $db->next_pos($status_cb) };
+    my $pos = eval { $db->next_pos };
 
     unless ($@) {
       $db->lock_un;
       defined($pos) or last;
-      $db->$status_cb;
+      $db->$status_cb(++$count);
       next;
     }
     warn $@;
 
     unless ($ptr_pos > 0) {
-      $db->warn("bad ptr $ptr_pos, cannot repair bucket $$db{cur_hash}");
+      $db->warn("bad ptr $ptr_ps, cannot repair bucket $$db{cur_hash}");
       next;
     }
 
-#    $db->seek($ptr_pos, SEEK_SET);
-#    my $pos = $db->read_int;
-#    $db->seek($pos, SEEK_SET);
-#    $db->lock_sh;
-#    my ($key, $next_pos) = eval { $db->read_key };
-#    $@ or $db->lock_un;
-#    $next_pos ||= 0;
-#    $next_pos == $ptr_pos - 1 and $next_pos = 0; # loops
-
-    my $next_pos = 0;
-    $db->warn("unlinking from $ptr_pos, to $next_pos (run defrag)");
+    $db->warn("unlinking from $ptr_pos, bucket $$db{cur_hash}");
     $db->seek($ptr_pos, SEEK_SET);
-    $db->write_int($next_pos);
+    $db->write_int(0);
   }
 
-  $db->sync;
   $db->lock_un;
 }
 
@@ -963,12 +889,11 @@ sub test {
   local $SIG{PIPE} = sub { };
   local $debug = 1;
   my $procs = 0;
-
-  ok 0, 65;
-
+  
   # clear
   $db->{hash_size} = 19;
   %db = ( );
+  ok 0, keys(%db) == 0;
 
   # store, fetch, delete, exists
   $db{hello} = 'world';
@@ -1005,8 +930,8 @@ sub test {
   for (1 .. 99) {
     wait, --$procs until $procs < $max_procs;
     ++$procs; fork and next;
-    my $key1 = 1 + int rand 49;
-    my $key2 = 51 + int rand 49;
+    my $key1 = 1 + int rand 100;
+    my $key2 = 1 + int rand 100;
     $db->lock_ex;
     @db{$key1, $key2} = @db{$key2, $key1};
     $db->lock_un;
@@ -1161,13 +1086,14 @@ sub test {
 
   $db->warn("warnings expected on test 38");
   $db->repair;
-  ok 38, keys(%db) <= 98;
+  my $keys = keys(%db);
+  ok 38, $keys < 100;
   ok 39, !exists $db{101}; 
   ok 40, !exists $db{99}; 
 
   $db->warn("warnings expected on test 41");
   $db->defrag;
-  ok 41, keys(%db) == 100;
+  ok 41, keys(%db) <= $keys;
 
   # no warnings
   my $keys = keys(%db);
@@ -1260,8 +1186,8 @@ sub test {
 
   ok 59, exists $db{$k1};
   ok 60, exists $db{$k2};
-  ok 61, exists $db{$k3};
-  ok 62, keys(%db) == $keys;
+  ok 61, !exists $db{$k3};
+  ok 62, keys(%db) < $keys;
 
   # ultimate test
   $db->warn('warnings expected on test 63');
